/* 

Zeitfunktionsgenerator 2 x 4Bit auf Basis 
Atmel ATtiny261 oder ATtiny261A (neuer)

http://www.gfai.de/~heinz/techdocs
heinz@gfai.de

Funktion

Auf PA0...3 und PA4...7 werden bei jedem Systick a 250s
zwei Nibble a 4 Bit ausgegeben. An diesen Portausgngen 
liegen zwei 4-Bit R-2R-Wandler, die als DAC je eine 
Zeitfunktion erzeugen. 

An drei Jumpern (JP1 PB4:6) knnen acht Zeitfunktionen eingestellt werden.
Der Systemtakt wird mit Timer0 eingestellt. Jeder Systick sorgt fr 
Abarbeitung des an den Jumpern eingestellten Zeitfunktionsprogramms. 

An Portpin PB3 ist eine LED angeschlossen, deren Blinkrate 
ber Timer1 gesteuert wird. 

PB0 wirkt als /STOP-Eingang: 
Solange PB0 = high (pull-up ist aktiv), luft Zhler cnt bis steps und 
beginnt von vorn.
Wenn PB0 = 0 wird, wird cnt in jedem Sample rckgesetzt, es wird keine 
Zeitfunktion ausgegeben. Siehe ISR(TIMER0_OVF_vect).

Digital-Analog-Wandler (DAC)
Die DAC besitzen eine Ausgangsimpedanz von etwa 1 kOhm. In der konkreten 
Applikation sind je drei 180 kOhm Widerstnde in Serie an den Ausgang 
geschaltet, um einen sehr geringen Ausgangsstrom (max. 8,8 A) zu erzeugen 
zur Imitation von Photodioden.
Die zwei R-2R-DAC knnnen im Prinzip auch ohne Timer betrieben werden, 
die Taktrate des Prozessors begrenzt dann deren maximale Arbeitsfrequenz.

Betriebsarten
Am PORTB 4,5,6 sind drei Jumper gegen Masse angeschlossen. 
Darber lassen sich 2^3 = 8 Betriebsarten whlen.
Jumper gesetzt erzeugt low am Portpin, die Jumper sind low-aktiv.
Das Programm cmdDecode() erhlt high-aktive Werte als Input. 
x heit input negiert, d.h. der Jumper auf diesem Pin ist gesetzt:

	JP1		Programm	Zeitfunktion
	654

	---		ZF_0	XXX...XXX	Dreiecksfolge
	--x		ZF_1	___/X\___	zwei Pyramiden
	-x-		ZF_2	___|X|___	nur pos. Halbwelle - Photodiodenchar.
	-xx		ZF_3	====XX===	Dreiecke inverse zueinander
	x--		ZF_4	___|X|___	pos. plus neg. Halbwelle
	x-x		ZF_5	_|=|-|=|_	Rechtecke
	xx-		ZF_6	X|X|X|X|X	Sge
	xxx		ZF_7	===|X====	Sgezahn

Variablen:
	cnt:	Zhler wird durch Timer0-Overflow-Interrupt inkrementiert
			und bei stepmax rckgesetzt
	lo:		Low-Nibble 	0x0_ <=> DAC0
	hi:		High-Nibble 0x_0 <=> DAC1

Programmierumgebung
AVR-Studio 4.19, Built 730, Atmel 2011 incl. WINAVR-GCC 

Inbetriebnahme
Zum Brennen sind alle AVR Programmieradapter im AVR-ISP-Mode geeignet. 
z.B. USBprog mit AVRISP-mkII (nur ber "Con"-Icon einstellbar)

Fuses Achtung
Der ATtiny261A hat standardmig die Fuse CLKDIV (/8) gesetzt, 
Fuse bitte rcksetzen. 
Ergebnis: 
	Extended	FF
	High		DF
	Low			E2

Funktionstest
Nach dem Brennen blinkt die LED je nach gesetztem Jumper verschieden schnell.
Wird /STOP (PB0) auf LOW gezogen, geht LED aus.

Betriebsspannung
Die Schaltung kann mit 5 Volt oder mit 3,3V betrieben werden.

Prozessorwahl
Der Prozessor ATtiny261 bzw. ATtiny261A (hier kompatibel)
ist zweimal im AVR-Studio-Projekt einzustellen:
1) unter "Project" /Configuration Options sowie 
2) unter dem Icon "AVR" /Main /Device

Die frhere Variante war ein #define:
#define __AVR_ATtiny261__	// Prozessor (ruft <avr/iotn261.h> ber io.h auf)
#define __AVR_ATtiny261A__	// Prozessor (ruft <avr/iotn261a.h> ber io.h auf)
*/


/********* Compiler *************************************************/

#define F_CPU 8000000UL			// 8 MHz interner RC-Oszillator
#include <avr/io.h>
#include <avr/interrupt.h>

// Systick:
#define systick32us		0x01	// 32 s
#define systick256us	0x02	// 256 s
#define systick2ms		0x03	// 2 ms
#define systick8ms		0x04	// 8 ms
#define systick32ms		0x05	// 32 ms

// Zyklusdauer:
#define stepmax 120			// Zyklusdauer, 120 ~ 30ms, 127 ~ 31,75ms, minimal 50

// PORTA
#define DAC0	0x0f			// DAC0 auf Bit PA0...3
#define DAC1	0xf0			// DAC1 auf Bit PA4...7

// PORTB
#define MOSI 	(1<<PIN0)		// AVR-ISP Programmer und Stop-Eingang
#define MISO 	(1<<PIN1)		// AVR-ISP Programmer
#define SCK 	(1<<PIN2)		// AVR-ISP Programmer
#define LED 	(1<<PIN3)		// grne LED
#define JP4 	(1<<PIN4)		// Jumper auf PB4
#define JP5 	(1<<PIN5)		// Jumper auf PB5
#define JP6 	(1<<PIN6)		// Jumper auf PB6
#define nRESET 	(1<<PIN7)		// AVR-ISP Programmer

// remember
#define bitset(PORT,BITNUM) ((PORT) |=  (1<<(BITNUM)))	// Set I/O bit; 	OR 		|=
#define bitclr(PORT,BITNUM) ((PORT) &=~ (1<<(BITNUM)))	// Clear I/O bit; 	AND 	&=
#define bittog(PORT,BITNUM) ((PORT) ^=  (1<<(BITNUM))) 	// Toggle I/O bit; 	XOR 	^=

// LED
void ledon(void)  { ((PORTB) |=  (LED));	}	// Set I/O bit; 	OR 		|=
void ledoff(void) { ((PORTB) &=~ (LED));	}	// Clear I/O bit; 	AND 	&=
void ledtog(void) { ((PORTB) ^=  (LED));	} 	// Toggle I/O bit; 	XOR 	^=

// globale Variablen
static uint8_t mode, lo, hi, cnt, steps;

/************ Inits ***************************************************/

void initPorts(void) {		// Init data directions and ports
	DDRA  = DAC0|DAC1;		// Datenrichtung 1: output; 0: input; PORTA ist DAC
	//PORTA = 0x00;			// Portbits setzen (bei in = 1: pullup)
	DDRB  = LED;			// LED ist Output = 1, Rest Input = 0 (PB7 = /RESET)
	PORTB = MOSI|JP4|JP5|JP6; 	// PB0,4,5,6 mit Pullup
};


void initSystickTimer(uint8_t prescal) {	// Init Timer0 Overflow-Interrupt
	// set prescaler TCCR0B 2:0 --> prescal = {1...5}
		//	value	bits	_bm		step		ovf after
		// 	/1: 	001 	0x01	125ns		 32s 
		//	/8: 	010		0x02	1s			256s
		//	/64: 	011		0x03	8s			 2,04ms
		//	/256: 	100		0x04	32s		 8,16ms
		//	/1024: 	101		0x05	0,128 ms	32,64ms
	TCCR0B |= prescal;	// set CS1 (/1024) 0,128*255 = 32,64ms
	// set TOIE1 overflow-interrupt enable
	TIMSK |= (1<<TOIE0);	// 0x02 = 0b0000 0010, Bit#1
};

void initLedClk(uint8_t beats) {	// Init Timer1 Overflow-Interrupt
	// set prescaler TCCR1B 3:0
	TCCR1B |= 0b00001111;	// CS1 = 0x0f ~ /16384, p.117, beatfreq 488 Hz = 2,048ms
	//TCCR1B |= 0b00001110;	// CS1 = 0x0e ~ /8192,  p.117, beatfreq 977 Hz = 1,024ms
	OCR1C = beats;			// 
	// set TOIE1 overflow-interrupt enable
	TIMSK |= (1<<TOIE1);	// 0x04 = 0b0000 0100, Bit#2
};

void disableTimers(void) {
	TIMSK &=~ (1<<TOIE0);	// 0x02 = 0b0000 0010, Bit#1
	TIMSK &=~ (1<<TOIE1);	// 0x04 = 0b0000 0100, Bit#2
}

void enableTimers(void) {
	TIMSK |= (1<<TOIE0);	// 0x02 = 0b0000 0010, Bit#1
	TIMSK |= (1<<TOIE1);	// 0x04 = 0b0000 0100, Bit#2
}


/*********************** Zeitfunktionen ******************************/


void ZF_0(void) {			// gegenlufig gezahnt	// XXX...XXXX
	steps = 32;				// Anzahl der Schritte im Zyklus
	if ((cnt & DAC0) < 16) {
		lo = (cnt & DAC0);
		hi = ((~cnt) << 4);	// negieren und vier Bit links schieben
	} 
	if (((cnt & 0x10) == 0x10)) {
		lo = ((~cnt) & DAC0);
		hi = ((cnt) << 4);	// negieren und vier Bit links schieben
	}
	// DAC-Ausgabe erfolgt zentral
}


void ZF_1(void) {			// zwei Pyramiden	___/X\___
	steps = 120;

	if (cnt > 47 ) {		// Rest nullen
		lo = 0;
		hi = 0;
	}
	if (cnt < 48) {
		lo = 0;
		hi = ((~cnt) << 4);	// 
	}
	if (cnt < 32) {
		lo = ((~cnt) & DAC0);	// negieren
		hi = (cnt << 4);		// vier Bit links schieben
	}
	if (cnt < 16) {
		lo = (cnt & DAC0);	// maskieren
		hi = 0;
	}
}


void ZF_2(void) { 	// gegenlufige Pulse auf positiver Halbwelle	___|X|___
	steps = 120;			// Anzahl der Schritte im Zyklus
	if (cnt < 8) {
		lo = cnt + 8;		// ab Mitte nach oben
		hi = ((~cnt) << 4);	// negieren und vier Bit links schieben
	}
	if (cnt == 8) {
		lo = 0x07;			// Mittenpegel einstellen
		hi = 0x70;			// Mittenpegel einstellen
	}
	// DAC-Ausgabe erfolgt zentral
}


void ZF_3(void) {			// Pyramide und Inverse 	===XX====
	steps = 120;			// Anzahl der Schritte im Zyklus
	if (cnt < 16) {
		lo = (cnt & DAC0);
		hi = ((~lo) << 4);	// negieren und vier Bit links schieben
	}
	else {
		if (cnt < 32) {
			lo = (~cnt & DAC0);
			hi = ((~lo) << 4);	// negieren und vier Bit links schieben
		}
	}
	// DAC-Ausgabe erfolgt zentral
}


void ZF_4(void) { 			// gegenlufige Pulse 	___|X|___
	steps = 120;			// Anzahl der Schritte im Zyklus
	if (cnt < 16) {
		lo = cnt;
		hi = ((~lo) << 4);	// negieren und vier Bit links schieben
	}
	if (cnt > 15) {
		lo = 0;
	}
	// DAC-Ausgabe erfolgt zentral
}


void ZF_5(void) {	// verschobene Rechteckfunktionen 	____|=|-|=|____
	steps = 120;			// Anzahl der Schritte im Zyklus
	if (cnt == 0) {
		lo = DAC0;
		hi = 0;
	}
	if (cnt == 8) {
		lo = DAC0;
		hi = DAC1;
	}
	if (cnt == 16) {
		lo = 0;
		hi = DAC1;
	}
	if (cnt == 24) {
		lo = 0;
		hi = 0;
	}
	// DAC-Ausgabe erfolgt zentral
}


void ZF_6(void) {			// gegenlufige Treppen 	X|X|X|X|X|
	steps = 32;				// Anzahl der Schritte im Zyklus
	lo = (cnt & DAC0);
	hi = ((~cnt << 4) & DAC1);
	// DAC-Ausgabe erfolgt zentral
}


void ZF_7(void) { 			// inverse Pulse 	===|X====
	steps = 120;			// Anzahl der Schritte im Zyklus
	if (cnt < 16) {
		lo = cnt;
		hi = ((~cnt) << 4);	// negieren und vier Bit links schieben
	}
	else {
		lo = 0;
		hi = 0;
	}
	// DAC-Ausgabe erfolgt zentral
}


/*****************************************************/


uint8_t checkJumper(void) {	// Jumper abfragen PB654, maskieren und invertieren
	// shift right 4 bit, PORTB bits 6,5,4 schieben auf 2,1,0
	// invertierend:
	return (~PINB >> 4) & 0x07;	// result bits {2,1,0}; return-values 0...7
	// nichtinvers:
	// return (PINB >> 4) & 0x07;	// result bits {2,1,0}; return-values 7...0
}

void cmdDecode(uint8_t number) {	// input: allowed values {0...7}
	switch(number) {
		case 0: {			// kein Jumper gesetzt "-", dabei sind PB654 high
			initLedClk(23);		// LED-Frequenz in ms * 2,048
			ZF_0();			// 
			break;
		}
		case 1: {			// Jumper PB4 gesetzt
			initLedClk(32);
			ZF_1();			// 
			break;
		}
		case 2: {			// Jumper PB5 gesetzt
			initLedClk(45);
			ZF_2();			// 
			break;
		}
		case 3: {			// Jumper PB4 und PB5 gesetzt
			initLedClk(64);
			ZF_3();			// 
			break;
		}
		case 4: {			// Jumper PB6 gesetzt
			initLedClk(90);
			ZF_4();			// 
			break;
		}
		case 5: {			// Jumper PB4 und PB6 gesetzt
			initLedClk(128);
			ZF_5();			// 
			break;
		}
		case 6: {			// Jumper PB5 und PB6 gesetzt
			initLedClk(181);
			ZF_6();			// 
			break;
		}
		case 7: {			// Jumper PB4, PB5, PB6 gesetzt, PB456 alle low
			initLedClk(255);
			ZF_7();			// 
			break;
		}
		default: {			// case 0
			mode = 0;
		}
	}
}


/*********************** Hauptprogramm ******************************/

int main(void) {
	steps = stepmax;
	initPorts();
	ledon();
	mode = checkJumper();
	initSystickTimer(systick256us);	// Abtastrate
	initLedClk(255);					// beats mal zwei Millisekunden
	sei();

	while (1) {
		asm volatile ("nop");
		// hier passiert nichts
		// den Rest machen die Interrupts 
	}

	return 0;
}


/*********************** Interrupts *********************************/


ISR(TIMER0_OVF_vect) {	// Interrupt Timer0 Overflow - Systick

	// Abschalten prfen, if cnt >= steps oder wenn STOP-Signal anliegt an PB0
	if ( (cnt >= steps) | !(PINB & (1 << PINB0)) ) {	// und wenn PB0 = 0 (STOP)
 		cnt = 0;		// Timer0 rcksetzen, von vorn
	}

	lo = lo & 0x0f;		// Wertebereich DAC0 beschneiden
	hi = hi & 0xf0;		// Wertebereich DAC1 beschneiden
	PORTA = (lo | hi);	// Werte zusammenfassen und ausgeben

	cmdDecode(mode);	// nchsten Wert berechnen
	cnt++;				// Zhler luft permanent um
}


ISR(TIMER1_OVF_vect) {	// Interrupt Timer1 Overflow - LED toggle
	if (!(PINB & (1 << PINB0)))		// wenn PB0 = 0 (STOP)
		ledoff();		// LED ausschalten
	else 	
		ledtog();		// blinken

	mode = checkJumper(); 	// Jumper JP1 abfragen
}




